@@ -12,10 +12,17 @@ let poly_compare=compare
 module RS = struct
   (* Poor man's splitter for version < 5.0                       *)
   (* This definition is shadowed by the [include] on OCaml >=5.0 *)
-  let split rs =
-    let bits = Random.State.bits rs in
-    let rs' = Random.State.make [|bits|] in
-    rs'
+  (* For the record, this is a hack:
+     Seeding a child RNG based on the output of a parent RNG
+     does not create an independent RNG. *)
+  (* copy of 4.14 Random.State.t to create a record of the right shape *)
+  type rs = { st : int array; mutable idx : int } [@@warning "-69"]
+  let split rs : Random.State.t =
+    let rs' = { st = Array.init 55 (fun _i -> Random.State.bits rs); idx = 0 } in
+    for i = 0 to 54 do
+      rs'.st.(i) <- (rs'.st.(i) lxor rs'.st.((i+1) mod 55)) land 0x3FFFFFFF;
+    done;
+    Obj.magic rs' (* sorry! *)
   include Random.State
   (* This is how OCaml 5.0 splits:       *)
   (* Split a new PRNG off the given PRNG *)
@@ -48,6 +55,19 @@ let _opt_sum a b = match a, b with
 
 let sum_int = List.fold_left (+) 0
 
+let rec list_split l len acc = match len,l with
+  | _,[]
+  | 0,_ -> List.rev acc, l
+  | _,x::xs -> list_split xs (len-1) (x::acc)
+
+let cut_exp_zero s =
+  match String.split_on_char 'e' s with
+  | [signif;exponent] ->
+    (match exponent.[0] with (* int_of_string removes a leading '0' *)
+     | '+' -> Printf.sprintf "%se+%i" signif (int_of_string exponent) (* keep a leading '+' *)
+     | _   -> Printf.sprintf "%se%i" signif (int_of_string exponent)) (* keep a leading '-' *)
+  | _ -> s
+
 exception Failed_precondition
 (* raised if precondition is false *)
 
@@ -79,6 +99,25 @@ module Seq = struct
 
   let cons x next () = Cons (x, next)
 
+  let rec force_drop n xs =
+    match xs() with
+    | Nil ->
+      Nil
+    | Cons (_, xs) ->
+      let n = n - 1 in
+      if n = 0 then
+        xs()
+      else
+        force_drop n xs
+
+  let drop n xs =
+    if n < 0 then invalid_arg "Seq.drop"
+    else if n = 0 then
+      xs
+    else
+      fun () ->
+        force_drop n xs
+
   (* End of copy of old functions. *)
 
   let is_empty (seq : _ t) : bool = match seq () with
@@ -235,6 +274,39 @@ module Tree = struct
   let rec applicative_take (n : int) (l : 'a t list) : 'a list t = match (n, l) with
     | (0, _) | (_, []) -> pure []
     | (n, (tree :: trees)) -> liftA2 List.cons tree (applicative_take (pred n) trees)
+
+  (** [drop_one l []] returns all versions of [l] with one element removed, for example
+      [drop_one [1;2;3] [] = [ [2;3]; [1;3]; [1;2] ]] *)
+  let rec drop_one (l : 'a list) (rev_prefix : 'a list) : 'a list list = match l with
+    | [] -> []
+    | x::xs -> (List.rev rev_prefix @ xs) :: drop_one xs (x::rev_prefix)
+
+  let rec build_list_shrink_tree (l : 'a t list) : 'a list t Seq.t = match l with
+    | [] -> Seq.empty
+    | _::_ ->
+      fun () ->
+        let len = List.length l in
+        if len < 4 then
+          let candidates = drop_one l [] in
+          List.fold_right (* try dropping each element in turn, starting with the list head *)
+            (fun cand acc -> Seq.cons (Tree (List.map root cand, build_list_shrink_tree cand)) acc)
+            candidates
+            (fun () -> children (sequence_list l) ()) ()  (* otherwise, reduce element(s) *)
+        else
+          let xs,ys = list_split l ((1 + len) / 2) [] in
+          let xs_roots = List.map root xs in
+          let ys_roots = List.map root ys in
+          (* Try reducing a list [1;2;3;4] in halves: [1;2] and [3;4] *)
+          Seq.cons (Tree (xs_roots, build_list_shrink_tree xs))
+            (Seq.cons (Tree (ys_roots, build_list_shrink_tree ys))
+               (fun () ->
+                  (* Try dropping an element from either half: [2;3;4] and [1;2;4] *)
+                  let rest = List.tl l in
+                  let rest_roots = List.map root rest in
+                  (Seq.cons (Tree (rest_roots, build_list_shrink_tree rest))
+                     (Seq.cons (Tree (xs_roots@(List.tl ys_roots), build_list_shrink_tree (xs@(List.tl ys))))
+                        (fun () -> children (sequence_list l) ())))  (* at bottom: reduce elements *)
+                    () )) ()
 end
 
 module Gen = struct
@@ -252,7 +324,11 @@ module Gen = struct
 
   let pure (a : 'a) : 'a t = fun _ -> Tree.pure a
 
-  let ap (f : ('a -> 'b) t) (x : 'a t) : 'b t = fun st -> Tree.ap (f st)  (x st)
+  let ap (f : ('a -> 'b) t) (x : 'a t) : 'b t = fun st ->
+    let st' = RS.split st in
+    let ftree = f st in
+    let xtree = x st' in
+    Tree.ap ftree xtree
 
   let (<*>) = ap
 
@@ -262,13 +338,26 @@ module Gen = struct
   let liftA3 (f : 'a -> 'b -> 'c -> 'd) (a : 'a t) (b : 'b t) (c : 'c t) : 'd t =
     (a >|= f) <*> b <*> c
 
+  let liftA4 (f : 'a -> 'b -> 'c -> 'd -> 'e) (a : 'a t) (b : 'b t) (c : 'c t) (d : 'd t) : 'e t =
+    (a >|= f) <*> b <*> c <*> d
+
+  let liftA5 (f : 'a -> 'b -> 'c -> 'd -> 'e -> 'f) (a : 'a t) (b : 'b t) (c : 'c t) (d : 'd t) (e : 'e t) : 'f t =
+    (a >|= f) <*> b <*> c <*> d <*> e
+
   let map2 = liftA2
 
   let map3 = liftA3
 
+  let map4 = liftA4
+
+  let map5 = liftA5
+
   let return = pure
 
-  let bind (gen : 'a t) (f : 'a -> ('b t)) : 'b t = fun st -> Tree.bind (gen st) (fun a -> f a st)
+  let bind (gen : 'a t) (f : 'a -> ('b t)) : 'b t = fun st ->
+    let st' = RS.split st in
+    let gentree = gen st in
+    Tree.bind gentree (fun a -> f a (RS.copy st'))
 
   let (>>=) = bind
 
@@ -316,10 +405,16 @@ module Gen = struct
     then Tree.Tree (true, Seq.return false_gen)
     else false_gen
 
-  let float : float t = fun st ->
-    let x = exp (RS.float st 15. *. (if RS.bool st then 1. else -1.))
-            *. (if RS.bool st then 1. else -1.)
-    in
+  let float : float t = fun st -> (* switch to [bits64] once lower bound reaches 4.14 *)
+    (* Technically we could write [15] but this is clearer *)
+    let four_bits_mask = 0b1111 in
+    (* Top 4 bits *)
+    let left = Int64.(shift_left (of_int (RS.bits st land four_bits_mask)) 60) in
+    (* Middle 30 bits *)
+    let middle = Int64.(shift_left (of_int (RS.bits st)) 30) in
+    (* Bottom 30 bits *)
+    let right = Int64.of_int (RS.bits st) in
+    let x = Int64.(float_of_bits (logor left (logor middle right))) in
     let shrink a = fun () -> Shrink.float_towards 0. a () in
     Tree.make_primitive shrink x
 
@@ -364,11 +459,18 @@ module Gen = struct
                    ~origin:(Option.value ~default:(pick_origin_within_range ~low ~high ~goal:0.) origin)
                    ~low
                    ~high in
+    let origin = origin -. low in
     (float_bound_inclusive ~origin (high -. low))
     >|= (fun x -> low +. x)
 
   let (--.) low high = float_range ?origin:None low high
 
+  let exponential (mean : float) =
+    if Float.is_nan mean then invalid_arg "Gen.exponential";
+    let unit_gen = float_bound_inclusive 1.0 in
+    map (fun p -> -. mean *. (log p)) unit_gen
+    (* See https://en.wikipedia.org/wiki/Relationships_among_probability_distributions *)
+
   let neg_int : int t = nat >|= Int.neg
 
   (** [option gen] shrinks towards [None] then towards shrinks of [gen]. *)
@@ -381,6 +483,12 @@ module Gen = struct
   (** [opt] is an alias of {!val:option} for backward compatibility. *)
   let opt = option
 
+  let result ?(ratio : float = 0.75) (ok_gen : 'a t) (err_gen : 'e t) : ('a, 'e) result t = fun st ->
+    let p = RS.float st 1. in
+    if p < (1. -. ratio)
+    then Tree.map (fun e -> Error e) (err_gen st)
+    else Tree.map (fun o -> Ok o) (ok_gen st)
+
   (* Uniform positive random int generator.
 
      We can't use {!RS.int} because the upper bound must be positive and is excluded,
@@ -484,13 +592,13 @@ module Gen = struct
   let (--) low high = int_range ?origin:None low high
 
   let oneof (l : 'a t list) : 'a t =
-    int_range 0 (List.length l - 1) >>= List.nth l
+    int_bound (List.length l - 1) >>= List.nth l
 
   let oneofl (l : 'a list) : 'a t =
-    int_range 0 (List.length l - 1) >|= List.nth l
+    int_bound (List.length l - 1) >|= List.nth l
 
   let oneofa (a : 'a array) : 'a t =
-    int_range 0 (Array.length a - 1) >|= Array.get a
+    int_bound (Array.length a - 1) >|= Array.get a
 
   (* NOTE: we keep this alias to not break code that uses [small_int]
      for sizes of strings, arrays, etc. *)
@@ -550,15 +658,32 @@ module Gen = struct
   (* A tail-recursive implementation over Tree.t *)
   let list_size (size : int t) (gen : 'a t) : 'a list t =
     fun st ->
+    let st' = RS.split st in
     Tree.bind (size st) @@ fun size ->
-    let rec loop n acc =
-      if n <= 0
-      then acc
-      else (loop [@tailcall]) (n - 1) (Tree.liftA2 List.cons (gen st) acc)
+    let st' = RS.copy st' in (* start each loop from same Random.State to recreate same element (prefix) *)
+    let rec loop n acc = (* phase 1: build a list of element trees, tail recursively *)
+      if n <= 0          (* phase 2: build a list shrink Tree of element trees, tail recursively *)
+      then List.fold_left (fun acc t -> Tree.liftA2 List.cons t acc) (Tree.pure []) acc
+      else (loop [@tailcall]) (n - 1) ((gen st')::acc)
     in
-    loop size (Tree.pure [])
+    loop size []
 
-  let list (gen : 'a t) : 'a list t = list_size nat gen
+  (** [list_ignore_size_tree] is a helper applying its own size shrinking heuristic,
+      and thus using only the root of [size]'s output shrink [Tree]. *)
+  let list_ignore_size_tree (size : int t) (gen : 'a t) : 'a list t = fun st ->
+    let st' = RS.split st in
+    let size = Tree.root (size st) in
+    let st' = RS.copy st' in (* start each loop from same Random.State to recreate same element (prefix) *)
+    let rec loop n acc = (* phase 1: build a list of element trees, tail recursively *)
+      if n <= 0          (* phase 2: build a list shrink Tree of element trees, tail recursively *)
+      then
+        let l = List.rev acc in
+        Tree.Tree (List.map Tree.root l, Tree.build_list_shrink_tree l)
+      else (loop [@tailcall]) (n - 1) ((gen st')::acc)
+    in
+    loop size []
+
+  let list (gen : 'a t) : 'a list t = list_ignore_size_tree nat gen
 
   let list_repeat (n : int) (gen : 'a t) : 'a list t = list_size (pure n) gen
 
@@ -674,12 +799,14 @@ module Gen = struct
 
   let bytes_size ?(gen = char) (size : int t) : bytes t = fun st ->
     let open Tree in
+    let st' = RS.split st in
     size st >>= fun size ->
     (* Adding char shrinks to a mutable list is expensive: ~20-30% cost increase *)
     (* Adding char shrinks to a mutable lazy list is less expensive: ~15% cost increase *)
+    let st' = RS.copy st' in (* start char generation from same Random.State to recreate same char prefix (when size shrinking) *)
     let char_trees_rev = ref [] in
     let bytes = Bytes.init size (fun _ ->
-                    let char_tree = gen st in
+                    let char_tree = gen st' in
                     char_trees_rev := char_tree :: !char_trees_rev ;
                     (* Performance: return the root right now, the heavy processing of shrinks can wait until/if there is a need to shrink *)
                     root char_tree) in
@@ -698,31 +825,45 @@ module Gen = struct
   let string_size ?(gen = char) (size : int t) : string t =
     bytes_size ~gen size >|= Bytes.unsafe_to_string
 
-  let bytes : bytes t = bytes_size nat
+  let bytes_of_char_list cs =
+    let b = Buffer.create (List.length cs) in
+    List.iter (fun c -> Buffer.add_char b c) cs;
+    let bytes = Buffer.to_bytes b in
+    Buffer.clear b;
+    bytes
 
-  let bytes_of gen = bytes_size ~gen nat
+  let bytes : bytes t = list char >|= bytes_of_char_list
 
-  let bytes_printable = bytes_size ~gen:printable nat
+  let bytes_of gen = list gen >|= bytes_of_char_list
 
-  let bytes_small st = bytes_size small_nat st
+  let bytes_printable = list printable >|= bytes_of_char_list
 
-  let bytes_small_of gen st = bytes_size ~gen small_nat st
+  let bytes_small = list_ignore_size_tree small_nat char >|= bytes_of_char_list
 
-  let string : string t = string_size nat
+  let bytes_small_of gen = list_ignore_size_tree small_nat gen >|= bytes_of_char_list
 
-  let string_of gen = string_size ~gen nat
+  let string_of_char_list cs =
+    let b = Buffer.create (List.length cs) in
+    List.iter (fun c -> Buffer.add_char b c) cs;
+    let str = Buffer.contents b in
+    Buffer.clear b;
+    str
 
-  let string_printable = string_size ~gen:printable nat
+  let string : string t = list char >|= string_of_char_list
 
-  let string_small st = string_size small_nat st
+  let string_of gen = list gen >|= string_of_char_list
 
-  let string_small_of gen st = string_size ~gen small_nat st
+  let string_printable = list printable >|= string_of_char_list
+
+  let string_small = list_ignore_size_tree small_nat char >|= string_of_char_list
+
+  let string_small_of gen = list_ignore_size_tree small_nat gen >|= string_of_char_list
 
   let small_string ?(gen=char) = string_small_of gen
 
-  let small_list gen = list_size small_nat gen
+  let small_list gen = list_ignore_size_tree small_nat gen
 
-  let small_array gen = array_size small_nat gen
+  let small_array gen = list_ignore_size_tree small_nat gen >|= Array.of_list
 
   let join (gen : 'a t t) : 'a t = gen >>= Fun.id
 
@@ -768,7 +909,7 @@ module Gen = struct
     make_primitive
       ~gen:(fun st -> gen st |> Tree.root)
       ~shrink
-  
+
   let no_shrink (gen: 'a t) : 'a t = set_shrink (fun _ -> Seq.empty) gen
 
   let (let+) = (>|=)
@@ -786,21 +927,30 @@ module Print = struct
   let unit _ = "()"
 
   let int = string_of_int
+  let int32 i = Int32.to_string i ^ "l"
+  let int64 i = Int64.to_string i ^ "L"
 
   let bool = string_of_bool
 
-  let float = string_of_float
-
-  let bytes = Bytes.to_string
+  let float f = (* Windows workaround to avoid leading exponent zero such as "-1.00001604579e-010" *)
+    if Sys.win32
+    then string_of_float f |> cut_exp_zero
+    else string_of_float f
 
   let string s = Printf.sprintf "%S" s
 
+  let bytes b = string (Bytes.to_string b)
+
   let char c = Printf.sprintf "%C" c
 
   let option f = function
     | None -> "None"
     | Some x -> "Some (" ^ f x ^ ")"
 
+  let result vp ep = function
+    | Error e -> "Error (" ^ ep e ^ ")"
+    | Ok v -> "Ok (" ^ vp v ^ ")"
+
   let pair a b (x,y) = Printf.sprintf "(%s, %s)" (a x) (b y)
 
   let triple a b c (x,y,z) = Printf.sprintf "(%s, %s, %s)" (a x) (b y) (c z)
@@ -973,6 +1123,8 @@ module Observable = struct
     let combine_f f s x = Hashtbl.seeded_hash s (f x)
 
     let int i = i land max_int
+    let int32 (i:int32) = Hashtbl.hash i
+    let int64 (i:int64) = Hashtbl.hash i
 
     let bool b = if b then 1 else 2
 
@@ -985,6 +1137,11 @@ module Observable = struct
     let option f = function
       | None -> 42
       | Some x -> combine 43 (f x)
+
+    let result vh eh = function
+      | Error e -> combine 17 (eh e)
+      | Ok v -> combine 19 (vh v)
+
     let list f l = List.fold_left (combine_f f) 0x42 l
 
     let array f l = Array.fold_left (combine_f f) 0x42 l
@@ -996,6 +1153,8 @@ module Observable = struct
     type 'a t = 'a -> 'a -> bool
 
     let int : int t = (=)
+    let int32 : int32 t = (=)
+    let int64 : int64 t = (=)
 
     let bytes : bytes t = (=)
 
@@ -1029,6 +1188,8 @@ module Observable = struct
       | None, Some _ -> false
       | Some x, Some y -> f x y
 
+    let result ok error r1 r2 = Result.equal ~ok ~error r1 r2
+
     let pair f g (x1,y1)(x2,y2) = f x1 x2 && g y1 y2
   end
 
@@ -1037,6 +1198,8 @@ module Observable = struct
   let bool : bool t = make ~hash:H.bool ~eq:Eq.bool Print.bool
 
   let int : int t = make ~hash:H.int ~eq:Eq.int Print.int
+  let int32 : int32 t = make ~hash:H.int32 ~eq:Eq.int32 Print.int32
+  let int64 : int64 t = make ~hash:H.int64 ~eq:Eq.int64 Print.int64
 
   let float : float t = make ~eq:Eq.float Print.float
 
@@ -1050,6 +1213,10 @@ module Observable = struct
     make ~hash:(H.option p.hash) ~eq:(Eq.option p.eq)
       (Print.option p.print)
 
+  let result op rp =
+    make ~hash:(H.result op.hash rp.hash) ~eq:(Eq.result op.eq rp.eq)
+      (Print.result op.print rp.print)
+
   let array p =
     make ~hash:(H.array p.hash) ~eq:(Eq.array p.eq) (Print.array p.print)
 
@@ -1150,14 +1317,14 @@ end = struct
       (* This only gets evaluated *after* the test was run for [tbl], meaning it is correctly
          populated with bindings recorded during the test already *)
       let current_bindings : (k * v Tree.t) list = List.rev !(root.p_tree_bindings_rev) in
-      let take_at_most_tree : int Tree.t = Tree.make_primitive (Shrink.int_towards 0) (List.length current_bindings) in
       let current_tree_bindings : (k * v) Tree.t list = List.map (fun (k, tree) -> Tree.map (fun v -> (k, v)) tree) current_bindings in
-      let shrunk_bindings_tree : (k * v) list Tree.t = Tree.bind take_at_most_tree (fun take_at_most -> Tree.applicative_take take_at_most current_tree_bindings) in
+      let shrunk_bindings_tree_seq : (k * v) list Tree.t Seq.t = Tree.build_list_shrink_tree current_tree_bindings in
       (* During shrinking, we don't want to record/add bindings, so [~extend:false]. *)
-      let shrunk_poly_tbl_tree : (k, v) t Tree.t = Tree.map (fun bindings -> List.to_seq bindings |> T.of_seq |> make ~extend:false) shrunk_bindings_tree in
-      (* [shrunk_poly_tbl_tree] is a bit misleading: its root *should* be the same as [root] but because of the required laziness
-         induced by the mutation of bindings, we don't use it, only graft its children to the original [root]. *)
-      Tree.children shrunk_poly_tbl_tree ()
+      let shrunk_poly_tbl_tree_seq : (k, v) t Tree.t Seq.t =
+        Seq.map (fun t -> Tree.map (fun bindings -> List.to_seq bindings |> T.of_seq |> make ~extend:false) t) shrunk_bindings_tree_seq in
+      (* [shrunk_poly_tbl_tree_seq] is a bit misleading: its head *should* be the same as [root] but because of the required laziness
+         induced by the mutation of bindings, we don't use it, only graft its tail to the original [root]. *)
+      Seq.drop 1 shrunk_poly_tbl_tree_seq ()
     in
     Tree.Tree (root, shrinks)
 
@@ -1169,7 +1336,6 @@ end
 (** Internal representation of functions, used for shrinking and printing (in case of error). *)
 type ('a, 'b) fun_repr_tbl = {
   fun_tbl: ('a, 'b) Poly_tbl.t; (** Input-output bindings *)
-  fun_gen: 'b Gen.t; (** How to generate output values *)
   fun_print: 'b Print.t option; (** How to print output values *)
   fun_default: 'b; (** Default value for all inputs not explicitly mapped in {!fun_tbl} *)
 }
@@ -1203,8 +1369,8 @@ module Fn = struct
 
   let make_ (r : 'a fun_repr) : 'a fun_ = Fun (r, function_of_repr r)
 
-  let mk_repr tbl gen ?print def =
-    Fun_tbl { fun_tbl=tbl; fun_gen=gen; fun_print=print; fun_default=def; }
+  let mk_repr tbl ?print def =
+    Fun_tbl { fun_tbl=tbl; fun_print=print; fun_default=def; }
 
   let map_repr f repr = Fun_map (f, repr)
 
@@ -1233,7 +1399,7 @@ module Fn = struct
   (** [gen_rep obs gen] creates a function generator. Input values are observed with [obs] and
       output values are generated with [gen]. *)
   let gen_rep (obs : 'a Observable.t) ?(print : 'b Print.t option) (gen : 'b Gen.t)  : ('a -> 'b) fun_repr Gen.t =
-    Gen.liftA2 (fun default_value poly_tbl -> mk_repr poly_tbl gen ?print default_value) gen (Poly_tbl.create ?v_print:print obs gen 8)
+    Gen.liftA2 (fun default_value poly_tbl -> mk_repr poly_tbl ?print default_value) gen (Poly_tbl.create ?v_print:print obs gen 8)
 
   let gen (obs : 'a Observable.t) ?(print : 'b Print.t option) (gen : 'b Gen.t)  : ('a -> 'b) fun_ Gen.t =
     Gen.map make_ (gen_rep obs gen ?print)
@@ -1730,6 +1896,7 @@ module Test = struct
     state.step state.test.name state.test input Failure;
     state.cur_max_fail <- state.cur_max_fail - 1;
     R.fail state.res ~steps ~msg_l input;
+    if state.cur_max_fail > 0 then CR_continue else
     CR_yield state.res
 
   (* [check_state state] applies [state.test] repeatedly ([iter] times)
@@ -1864,10 +2031,10 @@ module Test = struct
   let pp_print_test_fail name out l =
     let rec pp_list out = function
       | [] -> ()
-      | [x] -> Format.fprintf out "%s@," x
+      | [x] -> Format.pp_print_string out x
       | x :: y -> Format.fprintf out "%s@,%a" x pp_list y
     in
-    Format.fprintf out "@[test `%s`@ failed on ≥ %d cases:@ @[<v>%a@]@]"
+    Format.fprintf out "@[test `%s`@ failed on ≥ %d cases:@ @[<v>%a@]@]@,"
       name (List.length l) pp_list l
 
   let asprintf fmt =
@@ -1885,24 +2052,40 @@ module Test = struct
 
   let print_collect c =
     let out = Buffer.create 64 in
-    Hashtbl.iter
-      (fun case num -> Printf.bprintf out "%s: %d cases\n" case num) c;
+    let total,lab_len =
+      Hashtbl.fold
+        (fun case num (total,lab_len) -> (total+num, max lab_len (String.length case))) c (0,0) in
+    let sorted_cases = (* reverse compare yields decreasing order *)
+      Hashtbl.to_seq c |> List.of_seq |> List.sort (fun (_,n1) (_,n2) -> Int.compare n2 n1) in
+    List.iter
+      (fun (case,num) ->
+         let percentage = 100. *. (float num) /. (float total) in (* Workaround for Windows/Unix difference: *)
+         let percentage = (Float.round (10. *. percentage)) /. 10. in (* 100. *. 7525. /. 10000. -> 75.2 or 75.3 *)
+         let perc_str = Printf.sprintf "(%.1f%%)" percentage in
+         Printf.bprintf out " %-*s %6d cases %8s\n" (1+lab_len) (case ^ ":") num perc_str) sorted_cases;
     Buffer.contents out
 
   let stat_max_lines = 20 (* maximum number of lines for a histogram *)
 
   let print_stat ((name,_), tbl) =
-    let avg = ref 0. in
+    let neg_avg_summands = ref [] in
+    let pos_avg_summands = ref [] in
     let num = ref 0 in
     let min_idx, max_idx =
       Hashtbl.fold
         (fun i res (m1,m2) ->
-           avg := !avg +. float_of_int (i * res);
+           let avg_summand = float_of_int (i * res) in
+           if avg_summand < 0.
+           then neg_avg_summands := avg_summand::!neg_avg_summands
+             else pos_avg_summands := avg_summand::!pos_avg_summands;
            num := !num + res;
            min i m1, max i m2)
         tbl (max_int,min_int)
     in
-    (* compute average *)
+    (* compute average, summing positive/negative separately by order of magnitude *)
+    let neg_avg_sums = List.sort Float.compare !neg_avg_summands |> fun xs -> List.fold_right (+.) xs 0. in
+    let pos_avg_sums = List.sort Float.compare !pos_avg_summands |> List.fold_left (+.) 0. in
+    let avg = ref (neg_avg_sums +. pos_avg_sums) in
     if !num > 0 then (
       avg := !avg /. float_of_int !num
     );
@@ -1910,8 +2093,10 @@ module Test = struct
        https://en.wikipedia.org/wiki/Standard_deviation *)
     let stddev =
       Hashtbl.fold
-        (fun i res m -> m +. (float_of_int i -. !avg) ** 2. *. float_of_int res)
-        tbl 0.
+        (fun i res acc -> float_of_int res *. ((float_of_int i -. !avg) ** 2.) :: acc)
+        tbl []
+      |> List.sort Float.compare (* add summands in increasing order to preserve precision *)
+      |> List.fold_left (+.) 0.
       |> (fun s -> if !num>0 then s /. float_of_int !num else s)
       |> sqrt
     in
@@ -1952,10 +2137,13 @@ module Test = struct
          max_val := max !max_val new_count) tbl;
     (* print entries of the table, sorted by increasing index *)
     let out = Buffer.create 128 in
+    (* Windows workaround to avoid annoying exponent zero such as "1.859e+018" *)
+    let fmt_float f =
+      if f > 1e7 || f < -1e7 then cut_exp_zero (Printf.sprintf "%.3e" f) else Printf.sprintf "%.2f" f in
     Printf.bprintf out "stats %s:\n" name;
     Printf.bprintf out
-      "  num: %d, avg: %.2f, stddev: %.2f, median %d, min %d, max %d\n"
-      !num !avg stddev !median min_idx max_idx;
+      "  num: %d, avg: %s, stddev: %s, median %d, min %d, max %d\n"
+      !num (fmt_float !avg) (fmt_float stddev) !median min_idx max_idx;
     let indwidth =
       let str_width i = String.length (Printf.sprintf "%d" i) in
       List.map str_width [min_idx; max_idx; min_idx + bucket_size * hist_size] |> List.fold_left max min_int in
